//
// GameX
// Text Printer Class Code
// by Justin Pease
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include "puzzledemox-text.hpp"

// TextPrinter - Initialize
// Load text font image
void TextPrinter::Initialize(char * alphabet_image_path)
{
	text_img.Load(alphabet_image_path);
}

// Prints out a string of text graphically using text_img which was loaded from bmp
void TextPrinter::PrintText (int draw_mode, float font_size, char * str, int x_int, int y_int, ColorX & color, float angle, float scaley, bool rotate_baseline, bool fixed)
{
	if(font_size <= 0.0f) return;

	float x = (float)x_int;
	float y = (float)y_int;

	float cosAngle = cosf(angle*DEGtoRAD);
	float sinAngle = sinf(angle*DEGtoRAD);

	if(draw_mode & DRAWOP_CENTERED) {
		float xp = font_size/2.0f - GetVisibleStringWidth(str,font_size)/2.0f;
		if(angle!=0.0f && rotate_baseline) {
			x += xp * cosAngle;
			y += xp * sinAngle;
		} else {
			x += xp;
		}
	}
	
	float xi = x;
	float yi = y;
	float scale = font_size / 16.0f;
	scaley *= scale;
	int n = 0;
	while(str[n]) {
		int xs, ys; // x and y of source position (top-left)
		float kern;
		bool visible; // becomes false if this character is invisible or unknown and thus we don't have to draw it

		GetVisibleCharInfo(str,n,xs,ys,kern,visible);

		if(fixed) x += kern*font_size/16.0f/2;

		if(str[n] == 'g') y++;

		float xd, yd;
		if(rotate_baseline) {
			x -= xi;
			y -= yi;
			xd = x * cosAngle - y * sinAngle;
			yd = x * sinAngle + y * cosAngle;
			xd += xi;
			yd += yi;
			x += xi;
			y += yi;
		} else {
			xd = x;
			yd = y;
		}

		if(visible) {
			GameX.SetDrawAngle(-angle);
			GameX.SetDrawScale(scale, scaley);
			GameX.SetDrawShading(color);
			GameX.SetDrawPart(xs,ys,xs+16,ys+16);
			GameX.SetDrawMode(draw_mode);
			GameX.DrawImage(&text_img, (int)xd, (int)yd);
		}

		if(fixed)
			x += (font_size*15.0f - kern*font_size)/16.0f;
		else
			x += font_size - kern*font_size/16.0f;

		if(str[n] == 'g')
			y--; // reverse earlier vertical offset for letter g

		n++;
	}
}

float TextPrinter::GetVisibleCharWidth(char * str, int index, float font_size)
{
	return GetVisibleStringWidth(str,index,index,font_size);
}

float TextPrinter::GetVisibleStringWidth(char * str, float font_size)
{
	return GetVisibleStringWidth(str,0,32767,font_size);
}

float TextPrinter::GetVisibleStringWidth(char * str, int first_index, int last_index, float font_size)
{
	if(font_size <= 0.0f) return 0.0f;
	int xs,ys; float kern; bool visible;
	float width = 0.0f;
	for(int n = first_index ; n <= last_index && str[n] ; n++) {
		GetVisibleCharInfo(str,n,xs,ys,kern,visible);
		width += (font_size - kern*font_size/16.0f);
	}
	return width;
}

void TextPrinter::GetVisibleCharInfo(char * str, int n, int & xs, int & ys, float & kern, bool & visible)
{
	visible = true;
	kern = 7;
	switch(str[n]) {
	case '~':
		xs = 16*0;
		ys = 16*0;
		kern = 6;
		break;
	case '!':
		xs = 16*1;
		ys = 16*0;
		kern = 13;
		break;
	case '@':
		xs = 16*2;
		ys = 16*0;
		kern = 1;
		break;
	case '#':
		xs = 16*3;
		ys = 16*0;
		kern = 4;
		break;
	case '$':
		xs = 16*4;
		ys = 16*0;
		kern = 7;
		break;
	case '%':
		xs = 16*5;
		ys = 16*0;
		kern = 2;
		break;
	case '^':
		xs = 16*6;
		ys = 16*0;
		kern = 7;
		break;
	case '&':
		xs = 16*7;
		ys = 16*0;
		kern = 3;
		break;
	case '*':
		xs = 16*8;
		ys = 16*0;
		kern = 5;
		break;
	case '(':
		xs = 16*9;
		ys = 16*0;
		kern = 10;
		break;
	case ')':
		xs = 16*10;
		ys = 16*0;
		kern = 11;
		break;
	case '_':
		xs = 16*11;
		ys = 16*0;
		kern = 8;
		break;
	case '+':
		xs = 16*12;
		ys = 16*0;
		kern = 7;
		break;
	case ',':
		xs = 16*13;
		ys = 16*0;
		kern = 12;
		break;
	case '.':
		xs = 16*14;
		ys = 16*0;
		kern = 12;
		break;
	case '/':
		xs = 16*15;
		ys = 16*0;
		kern = 6;
		break;

	case '`':
		xs = 16*0;
		ys = 16*1;
		kern = 11;
		break;
	case '0':
		xs = 16*1;
		ys = 16*1;
		kern = 5;
		break;
	case '1':
		xs = 16*2;
		ys = 16*1;
		kern = 11;
		if(str[n+1] == '1')
			kern += 1;
		break;
	case '2':
		xs = 16*3;
		ys = 16*1;
		kern = 6;
		break;
	case '3':
		xs = 16*4;
		ys = 16*1;
		kern = 7;
		break;
	case '4':
		xs = 16*5;
		ys = 16*1;
		kern = 5;
		break;
	case '5':
		xs = 16*6;
		ys = 16*1;
		kern = 7;
		break;
	case '6':
		xs = 16*7;
		ys = 16*1;
		kern = 5;
		break;
	case '7':
		xs = 16*8;
		ys = 16*1;
		kern = 7;
		break;
	case '8':
		xs = 16*9;
		ys = 16*1;
		kern = 6;
		break;
	case '9':
		xs = 16*10;
		ys = 16*1;
		kern = 5;
		break;
	case '-':
		xs = 16*11;
		ys = 16*1;
		kern = 10;
		break;
	case '=':
		xs = 16*12;
		ys = 16*1;
		kern = 5;
		break;
	case ';':
		xs = 16*13;
		ys = 16*1;
		kern = 12;
		break;
	case ':':
		xs = 16*14;
		ys = 16*1;
		kern = 12;
		break;
	case '?':
		xs = 16*15;
		ys = 16*1;
		kern = 7;
		break;

	case 'A':
		xs = 16*0;
		ys = 16*2;
		kern = 4;
		break;
	case 'B':
		xs = 16*1;
		ys = 16*2;
		kern = 7;
		break;
	case 'C':
		xs = 16*2;
		ys = 16*2;
		kern = 4;
		break;
	case 'D':
		xs = 16*3;
		ys = 16*2;
		kern = 4;
		break;
	case 'E':
		xs = 16*4;
		ys = 16*2;
		kern = 6;
		break;
	case 'F':
		xs = 16*5;
		ys = 16*2;
		kern = 7;
		break;
	case 'G':
		xs = 16*6;
		ys = 16*2;
		kern = 4;
		break;
	case 'H':
		xs = 16*7;
		ys = 16*2;
		kern = 5;
		break;
	case 'I':
		xs = 16*8;
		ys = 16*2;
		kern = 12;
		break;
	case 'J':
		xs = 16*9;
		ys = 16*2;
		kern = 9;
		break;
	case 'K':
		xs = 16*10;
		ys = 16*2;
		kern = 6;
		break;
	case 'L':
		xs = 16*11;
		ys = 16*2;
		kern = 7;
		if(str[n+1] == 'v' || str[n+1] == ' ')
			kern += 4;
		break;
	case 'M':
		xs = 16*12;
		ys = 16*2;
		kern = 3;
		break;
	case 'N':
		xs = 16*13;
		ys = 16*2;
		kern = 5;
		break;
	case 'O':
		xs = 16*14;
		ys = 16*2;
		kern = 3;
		break;
	case 'P':
		xs = 16*15;
		ys = 16*2;
		kern = 7;
		break;

	case 'Q':
		xs = 16*0;
		ys = 16*3;
		kern = 1;
		break;
	case 'R':
		xs = 16*1;
		ys = 16*3;
		kern = 5;
		if(str[n+1] == 'a' || str[n+1] == 'e' || str[n+1] == 'o')
			kern++;
		break;
	case 'S':
		xs = 16*2;
		ys = 16*3;
		kern = 7;
		break;
	case 'T':
		xs = 16*3;
		ys = 16*3;
		kern = 3;
		if(str[n+1] == 'e' || str[n+1] == 'o' || str[n+1] == 'y')
			kern += 4;
		break;
	case 'U':
		xs = 16*4;
		ys = 16*3;
		kern = 6;
		break;
	case 'V':
		xs = 16*5;
		ys = 16*3;
		kern = 4;
		break;
	case 'W':
		xs = 16*6;
		ys = 16*3;
		kern = 1;
		break;
	case 'X':
		xs = 16*7;
		ys = 16*3;
		kern = 4;
		break;
	case 'Y':
		xs = 16*8;
		ys = 16*3;
		kern = 5;
		if(str[n+1] == 'e' || str[n+1] == 'o' || str[n+1] == '.' || str[n+1] == ',' || str[n+1] == ' ')
			kern += 2;
		break;
	case 'Z':
		xs = 16*9;
		ys = 16*3;
		kern = 5;
		break;
	case 'a':
		xs = 16*10;
		ys = 16*3;
		kern = 7;
		break;
	case 'b':
		xs = 16*11;
		ys = 16*3;
		kern = 6;
		break;
	case 'c':
		xs = 16*12;
		ys = 16*3;
		kern = 7;
		if(str[n+1] == 'e' || str[n+1] == 'o')
			kern++;
		break;
	case 'd':
		xs = 16*13;
		ys = 16*3;
		kern = 5;
		break;
	case 'e':
		xs = 16*14;
		ys = 16*3;
		kern = 7;
		break;
	case 'f':
		xs = 16*15;
		ys = 16*3;
		kern = 8;
		break;

	case 'g':
		xs = 16*0;
		ys = 16*4;
		kern = 5;
		break;
	case 'h':
		xs = 16*1;
		ys = 16*4;
		kern = 6;
		break;
	case 'i':
		xs = 16*2;
		ys = 16*4;
		kern = 12;
		break;
	case 'j':
		xs = 16*3;
		ys = 16*4;
		kern = 9;
		break;
	case 'k':
		xs = 16*4;
		ys = 16*4;
		kern = 6;
		if(str[n+1] == 'e' || str[n+1] == 'o' || str[n+1] == 'd' || str[n+1] == ' ')
			kern++;
		break;
	case 'l':
		xs = 16*5;
		ys = 16*4;
		kern = 12;
		break;
	case 'm':
		xs = 16*6;
		ys = 16*4;
		kern = 1;
		break;
	case 'n':
		xs = 16*7;
		ys = 16*4;
		kern = 7;
		break;
	case 'o':
		xs = 16*8;
		ys = 16*4;
		kern = 6;
		break;
	case 'p':
		xs = 16*9;
		ys = 16*4;
		kern = 6;
		break;
	case 'q':
		xs = 16*10;
		ys = 16*4;
		kern = 6;
		break;
	case 'r':
		xs = 16*11;
		ys = 16*4;
		kern = 9;
		if(str[n+1] == 'e' || str[n+1] == 'o' || str[n+1] == 'd' || str[n+1] == '.' || str[n+1] == ',' || str[n+1] == ' ')
			kern++;
		break;
	case 's':
		xs = 16*12;
		ys = 16*4;
		kern = 8;
		break;
	case 't':
		xs = 16*13;
		ys = 16*4;
		kern = 10;
		if(str[n+1] == 'u')
			kern--;
		break;
	case 'u':
		xs = 16*14;
		ys = 16*4;
		kern = 7;
		break;
	case 'v':
		xs = 16*15;
		ys = 16*4;
		kern = 5;
		break;

	case 'w':
		xs = 16*0;
		ys = 16*5;
		kern = 2;
		break;
	case 'x':
		xs = 16*1;
		ys = 16*5;
		kern = 6;
		break;
	case 'y':
		xs = 16*2;
		ys = 16*5;
		kern = 5;
		if(str[n+1] == 'e' || str[n+1] == 'o' || str[n+1] == '.' || str[n+1] == ',' || str[n+1] == ' ')
			kern += 1;
		break;
	case 'z':
		xs = 16*3;
		ys = 16*5;
		kern = 6;
		break;
	case '<':
		xs = 16*4;
		ys = 16*5;
		kern = 6;
		break;
	case '>':
		xs = 16*5;
		ys = 16*5;
		kern = 6;
		break;
	case '\\':
		xs = 16*6;
		ys = 16*5;
		kern = 7;
		break;
	case '|':
		xs = 16*7;
		ys = 16*5;
		kern = 12;
		break;
	case '':
		xs = 16*8;
		ys = 16*5;
		kern = 9;
		break;
	case '':
	case '"':
		xs = 16*9;
		ys = 16*5;
		kern = 8;
		break;
	case '':
		xs = 16*10;
		ys = 16*5;
		kern = 12;
		break;
	case '':
	case '\'':
		xs = 16*11;
		ys = 16*5;
		kern = 12;
		break;
	case '[':
		xs = 16*12;
		ys = 16*5;
		kern = 10;
		break;
	case ']':
		xs = 16*13;
		ys = 16*5;
		kern = 10;
		break;
	case '{':
		xs = 16*14;
		ys = 16*5;
		kern = 10;
		break;
	case '}':
		xs = 16*15;
		ys = 16*5;
		kern = 10;
		break;
	default:
		visible = false;
		kern = 8;
		break;
	}

	if(str[n+1] == 'A' && (str[n] == 'W' || str[n] == 'P' || str[n] == 'T' || str[n] == 'F' || str[n] == 'f' || str[n] == 'r' || str[n] == 'V' || str[n] == 'W' || str[n] == 'Y' || str[n] == '7' || str[n] == '/' || str[n] == '^' || str[n] == '"' || str[n] == '\'' || str[n] == '' || str[n] == '' || str[n] == '' || str[n] == ''))
		kern += 3;
}
